import asyncio
import random
import typing

import aiohttp
import discord

from . import exceptions
from .node import Node
from .player import Player


class Client:

    def __init__(self, bot, loop=None, session=None):

        self.bot = bot
        self.loop = loop or asyncio.get_event_loop()
        self.session = session or aiohttp.ClientSession(loop=self.loop)
        self.bot.add_listener(self.update_handler, "on_socket_response")

        self.nodes = {}

    def __repr__(self):
        return f"<GraniteClient node_count={len(self.nodes.values())} player_count={len(self.players.values())}>"

    @property
    def players(self):

        players = []
        for node in self.nodes.values():
            players.extend(node.players.values())

        return {player.guild.id: player for player in players}

    async def update_handler(self, data: dict):

        if not data:
            return

        if data["t"] == "VOICE_SERVER_UPDATE":

            guild_id = int(data["d"]["guild_id"])
            try:
                player = self.players[guild_id]
                await player.voice_server_update(data["d"])
            except KeyError:
                return

        elif data["t"] == "VOICE_STATE_UPDATE":

            if int(data["d"]["user_id"]) != self.bot.user.id:
                return

            guild_id = int(data["d"]["guild_id"])
            try:
                player = self.players[guild_id]
                await player.voice_state_update(data["d"])
            except KeyError:
                return

        else:
            return

    async def create_node(self, host: str, port: int, password: str, identifier: str):

        node = Node(client=self, bot=self.bot,
                    host=host, port=port,
                    password=password, identifier=identifier)
        await node.connect()

        return node

    def get_node(self):

        if not self.nodes:
            raise exceptions.NoNodesAvailable("There are no nodes available.")

        return random.choice([node for node in self.nodes.values()])

    def create_player(self, guild: discord.Guild, cls: typing.Type[Player]):

        if not self.nodes:
            raise exceptions.NoNodesAvailable("There are no nodes available.")

        if guild.id in self.players.keys():
            raise exceptions.PlayerAlreadyExists(f"A player for guild '{guild.id}' already exists.")

        node = self.get_node()
        player = cls(self.bot, node, guild)
        node.players[guild.id] = player
        return node.players[guild.id]

    def get_player(self, guild: discord.Guild, cls: typing.Type[Player] = Player):

        if not self.nodes:
            raise exceptions.NoNodesAvailable("There are no nodes available.")

        if guild.id not in self.players.keys():
            self.create_player(guild, cls)

        return self.players[guild.id]



class GranitepyException(Exception):
    """Base of all granitepy exceptions."""
    pass


class FilterInvalidArgument(GranitepyException):
    """An invalid argument was passed to a filter."""
    pass


class NodeConnectionFailure(GranitepyException):
    """There was a problem while connecting to the node."""
    pass


class NodeConnectionClosed(GranitepyException):
    """This node was closed."""
    pass


class NodeNotAvailable(GranitepyException):
    """This node is not currently available"""
    pass


class NoNodesAvailable(GranitepyException):
    """There are no nodes currently available"""
    pass


class GuildNotFound(GranitepyException):
    """The guild passed to a player was not found."""
    pass


class PlayerAlreadyExists(GranitepyException):
    """A player for this guild already exists."""
    pass


class TrackInvalidPosition(GranitepyException):
    """An invalid position was chosen for a track."""
    pass


class TrackLoadError(GranitepyException):
    """There was an error while loading a track."""
    pass



"""
Made by twitch#7443 and MrRandom#9258
"""
__version__ = "0.1.0a"

from .client import Client
from .player import Player
from .node import Node
from .exceptions import *
from .events import *
from .filters import *
from .objects import *




class Track:

    def __init__(self, track_id: str, info: dict):

        self.track_id = track_id
        self.info = info

        self.title = info.get("title")
        self.author = info.get("author")
        self.length = info.get("length")
        self.yt_id = info.get("identifier")
        self.uri = info.get("uri")
        self.is_stream = info.get("isStream")
        self.is_seekable = info.get("isSeekable")
        self.position = info.get("position")

    def __str__(self):
        return self.title

    def __repr__(self):
        return f"<GraniteTrack title={self.title!r} uri={self.uri!r} length={self.length}>"


class Playlist:

    def __init__(self, playlist_info: dict, tracks: list):

        self.playlist_info = playlist_info
        self.tracks_raw = tracks

        self.name = playlist_info.get("name")
        self.selected_track = playlist_info.get("selectedTrack")

        self.tracks = [Track(track_id=track["track"], info=track["info"]) for track in self.tracks_raw]

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<GranitePlaylist name={self.name!r} track_count={len(self.tracks)}>"


class Metadata:

    def __init__(self, metadata: dict):

        self.version = metadata.get("version", 0)
        self.version_major = metadata.get("versionMajor", 0)
        self.version_minor = metadata.get("versionMinor", 0)
        self.version_revision = metadata.get("versionRevision", 0)
        self.version_commit = metadata.get("versionCommit", 0)
        self.version_build = metadata.get("versionBuild", 0)
        self.node_region = metadata.get("nodeRegion")
        self.node_id = metadata.get("nodeId")
        self.enabled_sources = metadata.get("enabledSources")
        self.loaded_plugins = metadata.get("loadedPlugins")

    def __str__(self):
        return self.version

    def __repr__(self):
        return f"<GraniteMetadata version={self.version!r} region={self.node_region!r} id={self.node_id} enabled_sources={self.enabled_sources}>"



class GranitepyEvent:

    name = "null_event"


class TrackStartEvent(GranitepyEvent):

    def __init__(self, player, data):

        self.player = player
        self.name = "track_start"

        self.track = data["track"]


class TrackEndEvent(GranitepyEvent):

    def __init__(self, player, data):

        self.player = player
        self.name = "track_end"

        self.track = data["track"]
        self.reason = data["reason"]
        self.may_start_next = data["mayStartNext"]


class TrackStuckEvent(GranitepyEvent):

    def __init__(self, player, data):

        self.player = player
        self.name = "track_stuck"

        self.track = data["track"]
        self.threshold = data["thresholdMs"]


class TrackExceptionEvent(GranitepyEvent):

    def __init__(self, player, data):

        self.player = player
        self.name = "track_exception"

        self.error = data["error"]
        self.exception = data["exception"]


class WebSocketClosedEvent(GranitepyEvent):

    def __init__(self, player, data):

        self.player = player
        self.name = "websocket_closed"

        self.reason = data["reason"]
        self.code = data["code"]
        self.by_remote = data["byRemote"]



import json
import socket
import time

import websockets

from . import events
from . import exceptions
from . import objects


class Node:

    def __init__(self, **kwargs):

        self.client = kwargs.pop("client")
        self.bot = kwargs.pop("bot")
        self.host = kwargs.pop("host")
        self.port = kwargs.pop("port")
        self.password = kwargs.pop("password")
        self.identifier = kwargs.pop("identifier")

        self.websocket_uri = f"ws://{self.host}:{self.port}/websocket"
        self.rest_uri = f"http://{self.host}:{self.port}/"

        self.websocket = None
        self.available = False
        self.task = None

        self.connection_id = None
        self.metadata = None

        self.headers = {
            "Authorization": self.password,
            "Andesite-Resume-Id": self.connection_id,
            "User-Id": self.bot.user.id
        }

        self.players = {}

    def __repr__(self):
        return f"<GraniteNode player_count={len(self.players.keys())} available={self.available}>"

    async def connect(self):

        await self.bot.wait_until_ready()

        try:
            self.websocket = await websockets.connect(uri=self.websocket_uri, extra_headers=self.headers)

            self.client.nodes[self.identifier] = self
            self.task = self.bot.loop.create_task(self.listen())
            self.available = True

        except websockets.InvalidHandshake:
            raise exceptions.NodeConnectionFailure(f"The password for node '{self.identifier}' is invalid.")
        except websockets.InvalidURI:
            raise exceptions.NodeConnectionFailure(f"The URI for node '{self.identifier}' is invalid.")
        except socket.gaierror:
            raise exceptions.NodeConnectionFailure(f"The node '{self.identifier}' failed to connect.")

    async def disconnect(self):

        for player in self.players.copy().values():
            await player.destroy()

        await self.websocket.close()
        del self.client.nodes[self.identifier]
        self.available = False
        self.task.cancel()

    async def listen(self):

        while self.available is True:

            try:
                data = await self.websocket.recv()

            except websockets.ConnectionClosed:
                await self.disconnect()
                raise exceptions.NodeConnectionClosed(f"The connection to node '{self.identifier}' was closed.")

            if not data:
                return

            data = json.loads(data)

            op_code = data.get("op")
            if op_code == "pong":
                self.bot.dispatch("node_ping", time.time())
            elif op_code == "stats":
                self.bot.dispatch(f"node_stats", data["stats"])
            elif op_code == "metadata":
                self.metadata = objects.Metadata(data["data"])
            elif op_code == "connection-id":
                self.connection_id = data["id"]
            elif op_code == "event":
                await self.dispatch_event(data)
            elif op_code == "player-update":
                try:
                    player = self.players[int(data["guildId"])]
                    await player.update_state(data["state"])
                except KeyError:
                    continue

    async def dispatch_event(self, data: dict):

        try:
            player = self.players[int(data["guildId"])]
        except KeyError:
            return

        event = getattr(events, data["type"], None)
        event = event(player, data)

        self.bot.dispatch(f"granitepy_{event.name}", event)

    async def send(self, **data):

        if not self.available:
            raise exceptions.NodeNotAvailable(f"The node '{self.identifier}' is not currently available.")

        await self.websocket.send(json.dumps(data))

    async def get_tracks(self, query: str):

        async with self.client.session.get(url=f"{self.rest_uri}/loadtracks",
                                           params=dict(identifier=query),
                                           headers={"Authorization": self.password}) as response:
            data = await response.json()

        load_type = data.get("loadType")

        if not load_type:
            raise exceptions.TrackLoadError("There was an error while trying to load this track.")

        elif load_type == "LOAD_FAILED":
            raise exceptions.TrackLoadError(f"There was an error of severity '{data['severity']}' while loading tracks.\n\n{data['cause']}")

        elif load_type == "NO_MATCHES":
            return None

        elif load_type == "PLAYLIST_LOADED":
            return objects.Playlist(playlist_info=data["playlistInfo"], tracks=data["tracks"])

        elif load_type == "SEARCH_RESULT" or load_type == "TRACK_LOADED":
            return [objects.Track(track_id=track["track"], info=track["info"]) for track in data["tracks"]]

    @property
    async def ping(self):

        start_time = time.time()
        await self.send(op="ping")
        end_time = await self.bot.wait_for(f"node_ping")

        return (end_time - start_time) * 1000

    @property
    async def stats(self):

        await self.send(op="get-stats")
        node_stats = await self.bot.wait_for(f"node_stats")

        return node_stats



from . import exceptions


class Filter:

    def __init__(self):
        self.payload = None


class Karaoke(Filter):

    def __init__(self, *, level: float, mono_level: float, filter_band: float, filter_width: float):
        super().__init__()

        self.level = level
        self.mono_level = mono_level
        self.filter_band = filter_band
        self.filter_width = filter_width

        self.payload = {"karaoke": {"level": self.level,
                                    "monoLevel": self.mono_level,
                                    "filterBand": self.filter_band,
                                    "filterWidth": self.filter_width}}

    def __repr__(self):
        return f"<GraniteFilterKaraoke level={self.level} mono_level={self.mono_level} filter_band={self.filter_band} filter_width={self.filter_width}>"


class Timescale(Filter):

    def __init__(self, *, speed: float, pitch: float, rate: float):
        super().__init__()

        if speed < 0:
            raise exceptions.FilterInvalidArgument("Timescale speed must be more than 0.")
        if pitch < 0:
            raise exceptions.FilterInvalidArgument("Timescale pitch must be more than 0.")
        if rate < 0:
            raise exceptions.FilterInvalidArgument("Timescale rate must be more than 0.")

        self.speed = speed
        self.pitch = pitch
        self.rate = rate

        self.payload = {"timescale": {"speed": self.speed,
                                      "pitch": self.pitch,
                                      "rate": self.rate}}

    def __repr__(self):
        return f"<GraniteFilterTimescale speed={self.speed} pitch={self.pitch} rate={self.rate}>"


class Tremolo(Filter):

    def __init__(self, *, frequency: float, depth: float):
        super().__init__()

        if frequency < 0:
            raise exceptions.FilterInvalidArgument("Tremolo frequency must be more than 0.")
        if 0 > depth >= 1:
            raise exceptions.FilterInvalidArgument("Tremolo depth must be between 0 and 1.")

        self.frequency = frequency
        self.depth = depth

        self.payload = {"tremolo": {"frequency": self.frequency,
                                    "depth": self.depth}}

    def __repr__(self):
        return f"<GraniteFilterTremolo frequency={self.frequency} depth={self.depth}>"


class Vibrato(Filter):

    def __init__(self, *, frequency: float, depth: float):

        super().__init__()
        if 0 > frequency >= 14:
            raise exceptions.FilterInvalidArgument("Vibrato frequency must be between 0 and 14.")
        if 0 > depth >= 1:
            raise exceptions.FilterInvalidArgument("Vibrato depth must be between 0 and 1.")

        self.frequency = frequency
        self.depth = depth

        self.payload = {"vibrato": {"frequency": self.frequency,
                                    "depth": self.depth}}

    def __repr__(self):
        return f"<GraniteFilterVibrato frequency={self.frequency} depth={self.depth}>"



import time
import typing

import discord

from . import objects
from . import filters
from . import exceptions
from .node import Node


class Player:

    def __init__(self, bot, node: Node, guild: discord.Guild):

        self.bot = bot
        self.node = node
        self.guild = guild
        self.voice_channel = None
        self.text_channel = None

        self.voice_state = {}
        self.player_state = {}

        self.last_position = 0
        self.last_update = 0
        self.time = 0
        self.volume = 100
        self.paused = False
        self.filters = None

        self.current = None

    def __repr__(self):
        return f"<GranitePlayer is_connected={self.is_connected} is_playing={self.is_playing}>"

    @property
    def position(self):

        if self.current is None:
            return 0

        difference = (time.time() * 1000) - self.last_update
        return self.last_position + difference

    @property
    def is_connected(self):
        return self.voice_channel is not None

    @property
    def is_playing(self):
        return self.is_connected and self.current is not None

    @property
    def is_paused(self):
        return self.is_connected and self.paused is True

    async def update_state(self, state: dict):

        self.player_state = state

        self.last_update = time.time() * 1000
        self.last_position = state.get("position", 0)
        self.time = state.get("time", 0)
        self.volume = state.get("volume", 100)
        self.paused = state.get("paused", False)
        self.filters = state.get("filters", self.filters)

    async def voice_server_update(self, data: dict):

        self.voice_state.update({"event": data})

        if {"sessionId", "event"} == self.voice_state.keys():
            await self.node.send(op="voice-server-update", guildId=str(self.guild.id), **self.voice_state)

    async def voice_state_update(self, data: dict):

        self.voice_state.update({"sessionId": data["session_id"]})

        if data['channel_id'] is None:
            self.voice_channel = None
            self.voice_state.clear()
        else:
            self.voice_channel = self.bot.get_channel(int(data["channel_id"]))

        if {"sessionId", "event"} == self.voice_state.keys():
            await self.node.send(op="voice-server-update", guildId=str(self.guild.id), **self.voice_state)

    async def get_tracks(self, query: str):

        return await self.node.get_tracks(query)

    async def connect(self, voice_channel: discord.VoiceChannel):

        self.voice_channel = voice_channel

        ws = self.bot._connection._get_websocket(self.guild.id)
        await ws.voice_state(self.guild.id, str(voice_channel.id))

    async def stop(self):

        self.current = None

        await self.node.send(op="stop", guildId=str(self.guild.id))

    async def disconnect(self):

        await self.stop()
        self.voice_channel = None

        ws = self.bot._connection._get_websocket(self.guild.id)
        await ws.voice_state(self.guild.id, None)

    async def destroy(self):

        del self.node.players[self.guild.id]

        await self.disconnect()
        await self.node.send(op="destroy", guildId=str(self.guild.id))

    async def play(self, track: objects.Track, start_position: int = 0):

        await self.node.send(op="play",
                             guildId=str(self.guild.id),
                             track=track.track_id,
                             start=start_position)

        self.current = track
        return self.current

    async def seek(self, position: int):

        if position < 0 or position > self.current.length:
            raise exceptions.TrackInvalidPosition(f"Seek position must be between 0 and the track length")

        await self.node.send(op="seek",
                             guildId=str(self.guild.id),
                             position=position)

        return self.position

    async def set_pause(self, pause: bool):

        await self.node.send(op="pause", guildId=str(self.guild.id), pause=pause)

        self.paused = pause
        return self.is_paused

    async def set_volume(self, volume: typing.Union[int, float]):

        await self.node.send(op="volume",
                             guildId=str(self.guild.id),
                             volume=volume)

        self.volume = volume
        return self.volume

    async def set_filter(self, filter_type: filters.Filter):

        await self.node.send(op="filters", **filter_type.payload, guildId=str(self.guild.id))
        return filter_type

    async def set_timescale(self, *, speed: float = 1, pitch: float = 1, rate: float = 1):
        return await self.set_filter(filters.Timescale(speed=speed, pitch=pitch, rate=rate))

    async def set_karaoke(self, *, level: float = 1, mono_level: float = 1, filter_band: float = 220, filter_width: float = 100):
        return await self.set_filter(filters.Karaoke(level=level, mono_level=mono_level, filter_band=filter_band, filter_width=filter_width))

    async def set_tremolo(self, *, frequency: float = 2, depth: float = 0.5):
        return await self.set_filter(filters.Tremolo(frequency=frequency, depth=depth))

    async def set_vibrato(self, frequency: float = 2, depth: float = 0.5):
        return await self.set_filter(filters.Vibrato(frequency=frequency, depth=depth))



import setuptools


with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md") as f:
    long_description = f.read()


setuptools.setup(
    name="granitepy",
    version="0.2.0a",
    author="MrRandom#9258 and twitch 🔋#7443",
    url="https://github.com/iDevision/granitepy",
    licence="AGPL-3.0",
    description="A python client for the audio provider Andesite for use with discord.py.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['andesite', 'granitepy'],
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=requirements,
    python_requires='>=3.6',
    classifiers=[
        "Framework :: AsyncIO",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet"
    ]
)
